:root {
  --mpris-album-art-overlay: rgba(0, 0, 0, 0.55);
  --mpris-button-hover: rgba(0, 0, 0, 0.5);
  --mpris-album-art-icon-size: 96px;
  --mpris-album-art-shadow: 0px 0px 10px rgba(0, 0, 0, 0.75);
}

.widget-mpris {
  padding: 0;

  /* The parent to all players */
  .widget-mpris-player {
    margin: 16px 20px;
    border-radius: var(--border-radius);
    box-shadow: var(--mpris-album-art-shadow);

    .mpris-background {
      filter: blur(10px);
    }

    .mpris-overlay {
      padding: 16px;
      background-color: var(--mpris-album-art-overlay);

      button:hover {
        /* The media player buttons (play, pause, next, etc...) */
        background: var(--noti-bg-hover);
      }

      .widget-mpris-album-art {
        border-radius: var(--border-radius);
        box-shadow: var(--mpris-album-art-shadow);
        -gtk-icon-size: var(--mpris-album-art-icon-size);
      }

      .widget-mpris-title {
        font-weight: bold;
        font-size: 1.25rem;
      }

      .widget-mpris-subtitle {
        font-size: 1.1rem;
      }

      &>box>button {

        /* Change player control buttons */
        &:hover {
          background-color: var(--mpris-button-hover);
        }
      }
    }
  }

  &>box>button {
    /* Change player side buttons */
  }

  &>box>button:disabled {
    /* Change player side buttons insensitive */
  }
}
