/* Fallback for older CSS themes */
@define-color cc-bg rgba(46, 46, 46, 0.7);

@define-color noti-border-color rgba(255, 255, 255, 0.15);
@define-color noti-bg rgba(48, 48, 48, 0.8);
@define-color noti-bg-opaque rgb(48, 48, 48);
@define-color noti-bg-darker rgb(38, 38, 38);
@define-color noti-bg-hover rgb(56, 56, 56);
@define-color noti-bg-hover-opaque rgb(56, 56, 56);
@define-color noti-bg-focus rgba(68, 68, 68, 0.6);
@define-color noti-close-bg rgba(255, 255, 255, 0.1);
@define-color noti-close-bg-hover rgba(255, 255, 255, 0.15);

@define-color text-color rgb(255, 255, 255);
@define-color text-color-disabled rgb(150, 150, 150);

@define-color bg-selected rgb(0, 128, 255);
